import pyomo.environ as pyo

from ..utilities import get_set_t, get_hour_factors, get_nr_timesteps_averaged


def delete_all_balances(model):
    """
    Deletes all balances (required if they need to be reconstructed)

    :param model: pyomo model
    :return: pyomo model
    """

    if model.find_component("block_network_constraints"):
        model.del_component(model.block_network_constraints)
    if model.find_component("const_energybalance"):
        model.del_component(model.const_energybalance)
    if model.find_component("const_emissions_tot"):
        model.del_component(model.const_emissions_tot)
        model.del_component(model.const_emissions_neg)
        model.del_component(model.const_emissions_net)
    if model.find_component("block_costbalance"):
        model.del_component(model.block_costbalance)
    if model.find_component("const_npv"):
        model.del_component(model.const_npv)
    if model.find_component("const_emissions"):
        model.del_component(model.const_emissions)

    return model


def construct_network_constraints(model, config: dict):
    """
    Construct the network constraints to calculate nodal in- and outflow

    .. math::
      outflowToNetwork = \\sum(outflow \forall arcs at node)

    .. math::
      inflowFromNetwork = \\sum(inflow \forall arcs at node)\\

    :param model: pyomo model
    :param dict config: dict containing model information
    :return: pyomo model
    """

    def init_network_constraints(b_netw_const, period):
        """Pyomo rule to generate network constraint block"""

        b_period = model.periods[period]

        set_t = get_set_t(config, b_period)

        def init_netw_inflow(const, node, car, t):
            if car in b_period.node_blocks[node].set_carriers:
                return b_period.node_blocks[node].var_netw_inflow[t, car] == sum(
                    b_period.network_block[netw].var_inflow[t, car, node]
                    for netw in b_period.set_networks
                    if car in b_period.network_block[netw].set_netw_carrier
                )
            else:
                return pyo.Constraint.Skip

        b_netw_const.const_netw_inflow = pyo.Constraint(
            model.set_nodes, model.set_carriers, set_t, rule=init_netw_inflow
        )

        def init_netw_outflow(const, node, car, t):

            if car in b_period.node_blocks[node].set_carriers:
                return b_period.node_blocks[node].var_netw_outflow[t, car] == sum(
                    b_period.network_block[netw].var_outflow[t, car, node]
                    for netw in b_period.set_networks
                    if car in b_period.network_block[netw].set_netw_carrier
                )
            else:
                return pyo.Constraint.Skip

        b_netw_const.const_netw_outflow = pyo.Constraint(
            model.set_nodes, model.set_carriers, set_t, rule=init_netw_outflow
        )

        def init_netw_consumption(const, node, car, t):

            if (b_period.node_blocks[node].find_component("var_netw_consumption")) and (
                car in b_period.node_blocks[node].set_carriers
            ):
                return b_period.node_blocks[node].var_netw_consumption[t, car] == sum(
                    b_period.network_block[netw].var_consumption[t, car, node]
                    for netw in b_period.set_networks
                    if (b_period.network_block[netw].find_component("var_consumption"))
                    and car in b_period.network_block[netw].set_consumed_carriers
                )
            else:
                return pyo.Constraint.Skip

        b_netw_const.const_netw_consumption = pyo.Constraint(
            model.set_nodes, model.set_carriers, set_t, rule=init_netw_consumption
        )

    model.block_network_constraints = pyo.Block(
        model.set_periods, rule=init_network_constraints
    )

    return model


def construct_nodal_energybalance(model, config: dict):
    """
    Calculates the energy balance for each node and carrier

    .. math::
        outputFromTechnologies - inputToTechnologies + \\
        inflowFromNetwork - outflowToNetwork + \\
        imports - exports = demand - genericProductionProfile

    :param model: pyomo model
    :param dict config: dict containing model information
    :return: pyomo model
    """

    def init_energybalance(b_ebalance, period):
        b_period = model.periods[period]

        set_t = get_set_t(config, b_period)

        # Violation variables and costs
        if config["energybalance"]["violation"]["value"] > 0:
            b_period.var_violation = pyo.Var(
                set_t,
                model.set_carriers,
                model.set_nodes,
                domain=pyo.NonNegativeReals,
            )

        def init_energybalance(const, t, car, node):
            if car in b_period.node_blocks[node].set_carriers:
                node_block = b_period.node_blocks[node]
                tec_output = sum(
                    node_block.tech_blocks_active[tec].var_output_tot[t, car]
                    for tec in node_block.set_technologies
                    if car in node_block.tech_blocks_active[tec].set_output_carriers_all
                )

                tec_input = sum(
                    node_block.tech_blocks_active[tec].var_input_tot[t, car]
                    for tec in node_block.set_technologies
                    if car in node_block.tech_blocks_active[tec].set_input_carriers_all
                )

                netw_inflow = node_block.var_netw_inflow[t, car]

                netw_outflow = node_block.var_netw_outflow[t, car]

                if hasattr(node_block, "var_netw_consumption"):
                    netw_consumption = node_block.var_netw_consumption[t, car]
                else:
                    netw_consumption = 0

                import_flow = node_block.var_import_flow[t, car]

                export_flow = node_block.var_export_flow[t, car]

                if config["energybalance"]["violation"]["value"] > 0:
                    violation = b_period.var_violation[t, car, node]
                else:
                    violation = 0
                return (
                    tec_output
                    - tec_input
                    + netw_inflow
                    - netw_outflow
                    - netw_consumption
                    + import_flow
                    - export_flow
                    + violation
                    == node_block.para_demand[t, car]
                    - node_block.var_generic_production[t, car]
                )
            else:
                return pyo.Constraint.Skip

        b_ebalance.const_energybalance = pyo.Constraint(
            set_t, model.set_carriers, model.set_nodes, rule=init_energybalance
        )

        return b_ebalance

    model.block_energybalance = pyo.Block(model.set_periods, rule=init_energybalance)

    return model


def construct_global_energybalance(model, config):
    """
    Calculates the global energy balance for each carrier summed over all nodes

    .. math::
        outputFromTechnologies - inputToTechnologies + \\
        imports - exports = demand - genericProductionProfile

    :param model: pyomo model
    :param dict config: dict containing model information
    :return: pyomo model
    """

    def init_energybalance(b_ebalance, period):
        b_period = model.periods[period]

        set_t = get_set_t(config, b_period)

        # Violation variables and costs
        if config["energybalance"]["violation"]["value"] >= 0:
            b_period.var_violation = pyo.Var(
                set_t,
                model.set_carriers,
                model.set_nodes,
                domain=pyo.NonNegativeReals,
            )

        def init_energybalance_global(const, t, car):
            tec_output = sum(
                sum(
                    b_period.node_blocks[node]
                    .tech_blocks_active[tec]
                    .var_output_tot[t, car]
                    for tec in b_period.node_blocks[node].set_technologies
                    if car in b_period.node_blocks[node].set_carriers
                    and b_period.node_blocks[node]
                    .tech_blocks_active[tec]
                    .set_output_carriers_all
                )
                for node in model.set_nodes
            )

            tec_input = sum(
                sum(
                    b_period.node_blocks[node]
                    .tech_blocks_active[tec]
                    .var_input_tot[t, car]
                    for tec in b_period.node_blocks[node].set_technologies
                    if car in b_period.node_blocks[node].set_carriers
                    and b_period.node_blocks[node]
                    .tech_blocks_active[tec]
                    .set_input_carriers_all
                )
                for node in model.set_nodes
            )

            import_flow = sum(
                b_period.node_blocks[node].var_import_flow[t, car]
                for node in model.set_nodes
                if car in b_period.node_blocks[node].set_carriers
            )

            export_flow = sum(
                b_period.node_blocks[node].var_export_flow[t, car]
                for node in model.set_nodes
                if car in b_period.node_blocks[node].set_carriers
            )

            demand = sum(
                b_period.node_blocks[node].para_demand[t, car]
                for node in model.set_nodes
                if car in b_period.node_blocks[node].set_carriers
            )

            gen_prod = sum(
                b_period.node_blocks[node].var_generic_production[t, car]
                for node in model.set_nodes
                if car in b_period.node_blocks[node].set_carriers
            )

            if config["energybalance"]["violation"]["value"] > 0:
                violation = sum(
                    b_period.var_violation[t, car, node]
                    for node in model.set_nodes
                    if car in b_period.node_blocks[node].set_carriers
                )
            else:
                violation = 0

            return (
                tec_output - tec_input + import_flow - export_flow + violation
                == demand - gen_prod
            )

        model.set_used_carriers = pyo.Set(
            initialize=list(
                set().union(
                    *[
                        b_period.node_blocks[node].set_carriers
                        for node in model.set_nodes
                    ]
                )
            )
        )

        b_ebalance.const_energybalance = pyo.Constraint(
            set_t, model.set_used_carriers, rule=init_energybalance_global
        )

        return b_ebalance

    model.block_energybalance = pyo.Block(model.set_periods, rule=init_energybalance)

    return model


def construct_emission_balance(model, data):
    """
    Calculates the total postive and negative emissions as well as the net emissions

    .. math::
        E_{pos, tot} = E_{pos, technologies} + E_{pos, carriers} + E_{pos,
        networks}

    .. math::
        E_{neg, tot} = E_{neg, technologies} + E_{neg, carriers}

    .. math::
        E_{net} = E_{pos, tot} - E_{neg, tot}


    :param model: pyomo model
    :param data: DataHandle
    :return: pyomo model
    """
    config = data.model_config

    def init_emissionbalance(b_emissionbalance, period):
        b_period = model.periods[period]
        set_t = get_set_t(config, b_period)
        hour_factors = get_hour_factors(config, data, period)
        nr_timesteps_averaged = get_nr_timesteps_averaged(config)

        # calculate total emissions from technologies, networks and importing/exporting carriers
        def init_emissions_pos(const):
            from_technologies = sum(
                sum(
                    sum(
                        b_period.node_blocks[node]
                        .tech_blocks_active[tec]
                        .var_tec_emissions_pos[t]
                        * nr_timesteps_averaged
                        * hour_factors[t - 1]
                        for t in set_t
                    )
                    for tec in b_period.node_blocks[node].set_technologies
                )
                for node in model.set_nodes
            )
            from_carriers = sum(
                sum(
                    b_period.node_blocks[node].var_car_emissions_pos[t]
                    * nr_timesteps_averaged
                    * hour_factors[t - 1]
                    for t in set_t
                )
                for node in model.set_nodes
            )
            if not config["energybalance"]["copperplate"]["value"]:
                from_networks = sum(
                    sum(
                        sum(
                            b_period.network_block[netw].var_netw_emissions_pos[t, node]
                            * nr_timesteps_averaged
                            * hour_factors[t - 1]
                            for t in set_t
                        )
                        for node in model.set_nodes
                    )
                    for netw in b_period.set_networks
                )
            else:
                from_networks = 0
            return (
                from_technologies + from_carriers + from_networks
                == b_period.var_emissions_pos
            )

        b_emissionbalance.const_emissions_tot = pyo.Constraint(rule=init_emissions_pos)

        def init_emissions_neg(const):
            from_technologies = sum(
                sum(
                    sum(
                        b_period.node_blocks[node]
                        .tech_blocks_active[tec]
                        .var_tec_emissions_neg[t]
                        * nr_timesteps_averaged
                        * hour_factors[t - 1]
                        for t in set_t
                    )
                    for tec in b_period.node_blocks[node].set_technologies
                )
                for node in model.set_nodes
            )
            from_carriers = sum(
                sum(
                    b_period.node_blocks[node].var_car_emissions_neg[t]
                    * nr_timesteps_averaged
                    * hour_factors[t - 1]
                    for t in set_t
                )
                for node in model.set_nodes
            )
            return from_technologies + from_carriers == b_period.var_emissions_neg

        b_emissionbalance.const_emissions_neg = pyo.Constraint(rule=init_emissions_neg)

        b_emissionbalance.const_emissions_net = pyo.Constraint(
            expr=b_period.var_emissions_pos - b_period.var_emissions_neg
            == b_period.var_emissions_net
        )

        return b_emissionbalance

    model.block_emissionbalance = pyo.Block(
        model.set_periods, rule=init_emissionbalance
    )

    return model


def construct_import_costs(b_period, data, period: str):
    """
    Calculates the total import costs for an investment period

    .. math::
        C_{import} = \\sum(p_{t, import} * F_{t, import})

    :param b_period: pyomo block for period
    :param data: DataHandle
    :param str period: investment period to calculate import cost for
    :return: pyomo constraint
    """
    config = data.model_config

    set_t = get_set_t(config, b_period)
    hour_factors = get_hour_factors(config, data, period)
    nr_timesteps_averaged = get_nr_timesteps_averaged(config)

    def init_cost_import(const):
        return b_period.var_cost_imports == sum(
            sum(
                sum(
                    b_period.node_blocks[node].var_import_flow[t, car]
                    * b_period.node_blocks[node].para_import_price[t, car]
                    * nr_timesteps_averaged
                    * hour_factors[t - 1]
                    for car in b_period.node_blocks[node].set_carriers
                )
                for t in set_t
            )
            for node in b_period.node_blocks
        )

    return pyo.Constraint(rule=init_cost_import)


def construct_export_costs(b_period, data, period):
    """
    Calculates the total export costs for an investment period

    .. math::
        C_{export} = \\sum(p_{t, export} * F_{t, export})

    :param b_period: pyomo block for period
    :param data: DataHandle
    :param str period: investment period to calculate export cost for
    :return: pyomo constraint
    """
    config = data.model_config

    set_t = get_set_t(config, b_period)
    hour_factors = get_hour_factors(config, data, period)
    nr_timesteps_averaged = get_nr_timesteps_averaged(config)

    def init_cost_export(const):
        return b_period.var_cost_exports == -sum(
            sum(
                sum(
                    b_period.node_blocks[node].var_export_flow[t, car]
                    * b_period.node_blocks[node].para_export_price[t, car]
                    * nr_timesteps_averaged
                    * hour_factors[t - 1]
                    for car in b_period.node_blocks[node].set_carriers
                )
                for t in set_t
            )
            for node in b_period.node_blocks
        )

    return pyo.Constraint(rule=init_cost_export)


def construct_system_cost(model, data):
    """
    Aggregates costs per investment period

    - Total capex of technologies
    - Total capex of networks
    - Total opex of technologies
    - Total opex of networks
    - Total cost of technologies (sum of opex and capex)
    - Total cost of networks (sum of opex and capex)
    - Total import costs
    - Total export costs
    - Total costs from violations of the energy balance
    - Carbon costs and revenues
    - Total cost per investment period as a sum of technology, network, import,
      export, violation and carbon costs

    :param model: pyomo model
    :param dict config: dict containing model information
    :return: pyomo model
    """
    config = data.model_config

    def init_period_cost(b_period_cost, period):
        b_period = model.periods[period]
        set_t = get_set_t(config, b_period)
        hour_factors = get_hour_factors(config, data, period)
        nr_timesteps_averaged = get_nr_timesteps_averaged(config)

        # Capex Tecs
        def init_cost_capex_tecs(const):
            return b_period.var_cost_capex_tecs == sum(
                sum(
                    b_period.node_blocks[node].tech_blocks_active[tec].var_capex_tot
                    for tec in b_period.node_blocks[node].set_technologies
                )
                for node in model.set_nodes
            )

        b_period_cost.const_capex_tecs = pyo.Constraint(rule=init_cost_capex_tecs)

        # Capex Networks
        def init_cost_capex_netws(const):
            if not config["energybalance"]["copperplate"]["value"]:
                return b_period.var_cost_capex_netws == sum(
                    b_period.network_block[netw].var_capex
                    for netw in b_period.set_networks
                )
            else:
                return b_period.var_cost_capex_netws == 0

        b_period_cost.const_capex_netw = pyo.Constraint(rule=init_cost_capex_netws)

        # Opex Tecs
        def init_cost_opex_tecs(const):
            tec_opex_variable = sum(
                sum(
                    sum(
                        b_period.node_blocks[node]
                        .tech_blocks_active[tec]
                        .var_opex_variable_tot[t]
                        * nr_timesteps_averaged
                        * hour_factors[t - 1]
                        for tec in b_period.node_blocks[node].set_technologies
                    )
                    for t in set_t
                )
                for node in model.set_nodes
            )

            tec_opex_fixed = sum(
                sum(
                    b_period.node_blocks[node]
                    .tech_blocks_active[tec]
                    .var_opex_fixed_tot
                    for tec in b_period.node_blocks[node].set_technologies
                )
                for node in model.set_nodes
            )

            return b_period.var_cost_opex_tecs == tec_opex_fixed + tec_opex_variable

        b_period_cost.const_opex_tecs = pyo.Constraint(rule=init_cost_opex_tecs)

        # Opex Networks
        def init_cost_opex_netws(const):
            if not config["energybalance"]["copperplate"]["value"]:
                netw_opex_variable = sum(
                    sum(
                        b_period.network_block[netw].var_opex_variable[t]
                        * nr_timesteps_averaged
                        * hour_factors[t - 1]
                        for netw in b_period.set_networks
                    )
                    for t in set_t
                )
                netw_opex_fixed = sum(
                    b_period.network_block[netw].var_opex_fixed
                    for netw in b_period.set_networks
                )
                return (
                    b_period.var_cost_opex_netws == netw_opex_fixed + netw_opex_variable
                )
            else:
                return b_period.var_cost_opex_netws == 0

        b_period_cost.const_opex_netw = pyo.Constraint(rule=init_cost_opex_netws)

        # Total technology costs
        def init_cost_tecs(const):
            return (
                b_period.var_cost_tecs
                == b_period.var_cost_capex_tecs + b_period.var_cost_opex_tecs
            )

        b_period_cost.const_cost_tecs = pyo.Constraint(rule=init_cost_tecs)

        # Total network costs
        def init_cost_netw(const):
            return (
                b_period.var_cost_netws
                == b_period.var_cost_capex_netws + b_period.var_cost_opex_netws
            )

        b_period_cost.const_cost_netws = pyo.Constraint(rule=init_cost_netw)

        # Total import/export cost
        b_period_cost.const_cost_import = construct_import_costs(b_period, data, period)
        b_period_cost.const_cost_export = construct_export_costs(b_period, data, period)

        # Total violation cost
        def init_violation_cost(const):
            if config["energybalance"]["violation"]["value"] >= 0:
                return (
                    b_period.var_cost_violation
                    == sum(
                        sum(
                            sum(
                                b_period.var_violation[t, car, node]
                                * hour_factors[t - 1]
                                for t in set_t
                            )
                            for car in model.set_carriers
                        )
                        for node in model.set_nodes
                    )
                    * config["energybalance"]["violation"]["value"]
                )
            else:
                return b_period.var_cost_violation == 0

        b_period_cost.const_violation_cost = pyo.Constraint(rule=init_violation_cost)

        # Emission cost and revenues (if applicable)
        def init_carbon_revenue(const):
            revenue_carbon_from_technologies = sum(
                sum(
                    sum(
                        b_period.node_blocks[node]
                        .tech_blocks_active[tec]
                        .var_tec_emissions_neg[t]
                        * nr_timesteps_averaged
                        * hour_factors[t - 1]
                        * b_period.node_blocks[node].para_carbon_subsidy[t]
                        for t in set_t
                    )
                    for tec in b_period.node_blocks[node].set_technologies
                )
                for node in model.set_nodes
            )
            return revenue_carbon_from_technologies == b_period.var_carbon_revenue

        b_period_cost.const_revenue_carbon = pyo.Constraint(rule=init_carbon_revenue)

        def init_carbon_cost(const):
            cost_carbon_from_technologies = sum(
                sum(
                    sum(
                        b_period.node_blocks[node]
                        .tech_blocks_active[tec]
                        .var_tec_emissions_pos[t]
                        * nr_timesteps_averaged
                        * hour_factors[t - 1]
                        * b_period.node_blocks[node].para_carbon_tax[t]
                        for t in set_t
                    )
                    for tec in b_period.node_blocks[node].set_technologies
                )
                for node in model.set_nodes
            )
            cost_carbon_from_carriers = sum(
                sum(
                    b_period.node_blocks[node].var_car_emissions_pos[t]
                    * nr_timesteps_averaged
                    * b_period.node_blocks[node].para_carbon_tax[t]
                    * hour_factors[t - 1]
                    for t in set_t
                )
                for node in model.set_nodes
            )
            if not config["energybalance"]["copperplate"]["value"]:
                cost_carbon_from_networks = sum(
                    sum(
                        sum(
                            b_period.network_block[netw].var_netw_emissions_pos[t, node]
                            * nr_timesteps_averaged
                            * hour_factors[t - 1]
                            * b_period.node_blocks[node].para_carbon_tax[t]
                            for t in set_t
                        )
                        for node in model.set_nodes
                    )
                    for netw in b_period.set_networks
                )
            else:
                cost_carbon_from_networks = 0
            return (
                cost_carbon_from_technologies
                + cost_carbon_from_carriers
                + cost_carbon_from_networks
                == b_period.var_carbon_cost
            )

        b_period_cost.const_cost_carbon = pyo.Constraint(rule=init_carbon_cost)

        def init_total_cost(const):
            return (
                b_period.var_cost_tecs
                + b_period.var_cost_netws
                + b_period.var_cost_imports
                + b_period.var_cost_exports
                + b_period.var_cost_violation
                + b_period.var_carbon_cost
                - b_period.var_carbon_revenue
                == b_period.var_cost_total
            )

        b_period_cost.const_cost = pyo.Constraint(rule=init_total_cost)

        return b_period_cost

    model.block_costbalance = pyo.Block(model.set_periods, rule=init_period_cost)

    return model


def construct_global_balance(model):
    """
    Calculates total npv and total emissions over all investment periods
    :param model: pyomo model
    :return: pyomo model
    """

    def init_npv(const):
        return (
            sum(model.periods[period].var_cost_total for period in model.set_periods)
            == model.var_npv
        )

    model.const_npv = pyo.Constraint(rule=init_npv)

    def init_emissions(const):
        return (
            sum(model.periods[period].var_emissions_net for period in model.set_periods)
            == model.var_emissions_net
        )

    model.const_emissions = pyo.Constraint(rule=init_emissions)

    return model



from pyomo.environ import Set, RangeSet, Var
import logging

log = logging.getLogger(__name__)


def construct_investment_period_block(b_period, data: dict):
    """
    SETS
    - set_networks: Set of networks for investment period
    - set_t_full: full set_t
    - set_t_clustered: clustered set_t (can be equal to set_t_full)

    VARIABLES

    Cost Variables

    - var_cost_capex_tecs: Total Capex of technologies for respective investment period
    - var_cost_capex_netws: Total Capex of networks for respective investment period
    - var_cost_opex_tecs: Total Opex (fixed and variable) of technologies for respective investment period
    - var_cost_opex_netws: Total Opex (fixed and variable) of networks for respective investment period
    - var_cost_tecs: Total technology costs
    - var_cost_netws: Total network costs
    - var_cost_imports: Total import costs
    - var_cost_exports: Total export costs
    - var_cost_violation: Total violation cost
    - var_carbon_revenue: Total carbon revenues from negative emission technologies
    - var_carbon_cost: Total carbon cost from technologies, networks and imports/exports
    - var_cost_total: Total annualized cost for respective investment period

    Emission Variables

    - var_emissions_pos: Positive emissions from technologies, networks and imports/exports
    - var_emissions_neg: Negative emissions from technologies and imports/exports
    - var_emissions_net: Net emissions in investment period

    :param b_period: pyomo block with investment period
    :param dict data: data containing model configuration
    :return: pyomo block with investment period
    """

    # PREPROCESSING
    investment_period = b_period.index()
    config = data["config"]
    topology = data["topology"]
    network_data = data["network_data"]

    # LOG
    log_msg = f"Constructing Investment Period {investment_period}"
    log.info(log_msg)

    # SETS
    b_period.set_networks = Set(initialize=list(network_data.keys()))

    # TIME PERIODS
    if config["optimization"]["typicaldays"]["N"]["value"] == 0:
        # No clustering
        if config["optimization"]["timestaging"]["value"] == 0:
            # no averaging
            b_period.set_t_full = RangeSet(1, len(topology["time_index"]["full"]))
            b_period.set_t_clustered = RangeSet(1, len(topology["time_index"]["full"]))
        else:
            # first stage averaging
            b_period.set_t_full = RangeSet(1, len(topology["time_index"]["averaged"]))
            b_period.set_t_clustered = RangeSet(
                1, len(topology["time_index"]["averaged"])
            )

    else:
        # Method 1 and 2
        b_period.set_t_full = RangeSet(1, len(topology["time_index"]["full"]))
        b_period.set_t_clustered = RangeSet(1, len(topology["time_index"]["clustered"]))

    # VARIABLES
    b_period.var_cost_capex_tecs = Var()
    b_period.var_cost_capex_netws = Var()
    b_period.var_cost_opex_tecs = Var()
    b_period.var_cost_opex_netws = Var()
    b_period.var_cost_tecs = Var()
    b_period.var_cost_netws = Var()
    b_period.var_cost_imports = Var()
    b_period.var_cost_exports = Var()
    b_period.var_cost_violation = Var()
    b_period.var_carbon_revenue = Var()
    b_period.var_carbon_cost = Var()
    b_period.var_cost_total = Var()

    b_period.var_emissions_pos = Var()
    b_period.var_emissions_neg = Var()
    b_period.var_emissions_net = Var()

    log_msg = f"Constructing Investment Period {investment_period} completed"
    log.warning(log_msg)

    return b_period



from ..components.utilities import perform_disjunct_relaxation


def construct_technology_block(b_tec, data: dict, set_t_full, set_t_clustered):
    """
    Construct technology block and performs disjunct relaxation if required

    :param b_tec: pyomo block with technology model
    :param dict data: data containing model configuration
    :param set_t_full: pyomo set containing timesteps
    :param set_t_clustered: pyomo set containing clustered timesteps
    :return: pyomo block with technology model
    """
    tec = b_tec.index()
    technology = data["technology_data"][tec]
    b_tec = technology.construct_tech_model(b_tec, data, set_t_full, set_t_clustered)
    if technology.big_m_transformation_required:
        b_tec = perform_disjunct_relaxation(b_tec)

    return b_tec



from pathlib import Path
import pandas as pd
import pvlib
import os
import json

from ..components.technologies import *

import logging

log = logging.getLogger(__name__)


def calculate_dni(data: pd.DataFrame, lon: float, lat: float) -> pd.Series:
    """
    Calculate direct normal irradiance from ghi and dhi

    :param pd.DataFrame data: climate data with columns ghi and dhi
    :param float lon: longitude
    :param float lat: latitude
    :return data: climate data including dni
    :rtype: pd.Series
    """
    zenith = pvlib.solarposition.get_solarposition(data.index, lat, lon)
    data["dni"] = pvlib.irradiance.dni(
        data["ghi"].to_numpy(), data["dhi"].to_numpy(), zenith["zenith"].to_numpy()
    )
    data["dni"] = data["dni"].fillna(0)
    data["dni"] = data["dni"].where(data["dni"] > 0, 0)

    return data["dni"]


def select_technology(tec_data: dict):
    """
    Returns the correct subclass for a technology

    :param dict tec_data: dictonary derived from the technology json files
    :return: Technology Class
    """
    # Generic tecs
    if tec_data["tec_type"] == "RES":
        return Res(tec_data)
    elif tec_data["tec_type"] == "CONV1":
        return Conv1(tec_data)
    elif tec_data["tec_type"] == "CONV2":
        return Conv2(tec_data)
    elif tec_data["tec_type"] == "CONV3":
        return Conv3(tec_data)
    elif tec_data["tec_type"] == "CONV4":
        return Conv4(tec_data)
    elif tec_data["tec_type"] == "STOR":
        return Stor(tec_data)
    elif tec_data["tec_type"] == "SINK":
        return Sink(tec_data)
    # Specific tecs
    elif tec_data["tec_type"] == "DAC_Adsorption":
        return DacAdsorption(tec_data)
    elif tec_data["tec_type"].startswith("GasTurbine"):
        return GasTurbine(tec_data)
    elif tec_data["tec_type"].startswith("HeatPump"):
        return HeatPump(tec_data)
    elif tec_data["tec_type"] == "HydroOpen":
        return HydroOpen(tec_data)
    elif tec_data["tec_type"] == "CCPP":
        return CCPP(tec_data)


def read_tec_data(tec_name: str, load_path: Path):
    """
    Loads the technology data from load_path and preprocesses it.

    :param str tec_name: technology name
    :param Path load_path: load path
    :param pd.DataFrame climate_data: Climate Data
    :param dict location: Dictonary with node location
    :return: Technology Class
    """
    tec_data = open_json(tec_name, load_path)
    tec_data["name"] = tec_name
    tec_data = select_technology(tec_data)

    # CCS
    if tec_data.component_options.ccs_possible:
        tec_data.ccs_data = open_json(tec_data.component_options.ccs_type, load_path)
    return tec_data


def open_json(tec: str, load_path: Path) -> dict:
    """
    Loops through load_path and subdirectories and returns json with name tec + ".json"

    :param str tec: name of technology to read json for
    :param Path load_path: directory path to loop through all subdirectories and search for tec + ".json"
    :return: Dictionary containing the json data
    :rtype: dict
    """
    # Read in JSON files
    for path, subdirs, files in os.walk(load_path):
        if "data" in locals():
            break
        else:
            for name in files:
                if (tec + ".json") == name:
                    filepath = os.path.join(path, name)
                    with open(filepath) as json_file:
                        data = json.load(json_file)
                    break

    # Assign name
    if "data" in locals():
        data["Name"] = tec
    else:
        raise Exception("There is no json data file for technology " + tec)

    return data


def check_input_data_consistency(path: Path):
    """
    Checks if the topology is consistent with the input data.

    Checks for:
    - is there a folder for each investment period?
    - is there a network file for each network defined?
    - are there all required files for all networks in the directory?
    - are node directories there?
    - is ClimateData, CarbonCost for each node there?
    - is Technologies.json there?
    - is there a json file for all technologies?
    - is there a carrier file for each defined carrier?

    :param Path path: path to check for consistency
    """

    def check_path_existance(path: Path, error_message: str):
        if not os.path.exists(path):
            raise Exception(error_message)

    # Convert to Path
    if isinstance(path, str):
        path = Path(path)

    # Read topology
    with open(path / "Topology.json") as json_file:
        topology = json.load(json_file)

    for investment_period in topology["investment_periods"]:

        # Check investment periods
        check_path = path / investment_period
        check_path_existance(
            check_path,
            f"The investment period {investment_period} is missing in {check_path}",
        )

        # Check networks
        check_path_existance(
            check_path / "Networks.json",
            f"A Network.json file is missing in {check_path}",
        )
        with open(check_path / "Networks.json") as json_file:
            all_networks = json.load(json_file)
        for type in all_networks.keys():
            networks = all_networks[type]
            for network in networks:
                check_path_existance(
                    check_path / "network_data" / (network + ".json"),
                    f"A json file for {network} is missing in {check_path / 'network_data'}",
                )
                check_path_existance(
                    check_path / "network_topology" / type,
                    f"A directory for {network} is missing in {check_path / 'network_topology'}",
                )
                check_path_existance(
                    check_path / "network_topology" / type / network / "connection.csv",
                    f"A connection.csv for {network} is missing in {check_path / 'network_topology' / type / network}",
                )
                check_path_existance(
                    check_path / "network_topology" / type / network / "distance.csv",
                    f"A distance.csv for {network} is missing in {check_path / 'network_topology' / type / network}",
                )

                if type == "existing":
                    check_path_existance(
                        check_path / "network_topology" / type / network / "size.csv",
                        f"A size.csv for {network} is missing in {check_path / 'network_topology' / type / network}",
                    )

        for node in topology["nodes"]:

            # Check nodes
            check_node_path = path / investment_period / "node_data" / node
            check_path_existance(
                check_node_path, f"The node {node} is missing in {check_node_path}"
            )

            # Check if all files are there
            check_path_existance(
                check_node_path / "ClimateData.csv",
                f"ClimateData.csv is missing in {check_node_path}",
            )
            check_path_existance(
                check_node_path / "CarbonCost.csv",
                f"CarbonCost.csv is missing in {check_node_path}",
            )
            check_path_existance(
                check_node_path / "Technologies.json",
                f"Technologies.json is missing in {check_node_path}",
            )

            # Check if all technologies have a json file
            with open(check_node_path / "Technologies.json") as json_file:
                technologies_at_node = json.load(json_file)
            technologies_at_node = set(
                list(technologies_at_node["existing"].keys())
                + technologies_at_node["new"]
            )
            for technology in technologies_at_node:
                check_path_existance(
                    check_node_path / "technology_data" / (technology + ".json"),
                    f"A json file for {technology} is missing in {check_node_path / 'technology_data'}",
                )
                # TODO: Check if carriers are in carrier set

            # Check if all carriers are there
            for carrier in topology["carriers"]:
                check_path_existance(
                    check_node_path / "carrier_data" / (carrier + ".csv"),
                    f"Data for carrier {carrier} is missing in {check_node_path}",
                )

    # Read config
    with open(path / "ConfigModel.json") as json_file:
        config = json.load(json_file)

    # Check that averaging and k-means is not used at same time
    if (config["optimization"]["typicaldays"]["N"]["value"] != 0) and (
        config["optimization"]["timestaging"]["value"] != 0
    ):
        raise Exception(
            "Using time step averaging and k-means clustering at the same"
            " time is not allowed"
        )

    log_msg = "Input data folder has been checked successfully - no errors occurred."
    log.info(log_msg)



import h5py
import numpy as np
import pandas as pd
from pathlib import Path


def print_h5_tree(file_path: Path | str):
    """
    Function to print the structure of a h5 file

    The structure of a h5 file is a tree structure: the h5 file is the root group,
    from which all groups stem, and datasets are the leaves contained within a
    group.

    :param Path, str file_path: Path to H5 File
    """
    with h5py.File(file_path, "r") as hdf_file:

        def print_attrs(name, obj):
            print(name)
            for key, val in obj.attrs.items():
                print(f"    {key}: {val}")

        hdf_file.visititems(print_attrs)


def extract_datasets_from_h5group(group, prefix: tuple = ()) -> dict:
    """
    Extracts datasets from a group within a h5 file

    Gets all datasets from a group of a h5 file and writes it to a multi-index
    dataframe using a recursive function

    :param group: froup of h5 file
    :param tuple prefix: required to search through the structure of the h5 tree if there are multiple subgroups in the
     group you specified, empty by default meaning it starts searching from the group specified.
    :return: dataframe containing all datasets in group
    :rtype: pd.DataFrame
    """
    data = {}
    for key, value in group.items():
        if isinstance(value, h5py.Group):
            data.update(extract_datasets_from_h5group(value, prefix + (key,)))
        elif isinstance(value, h5py.Dataset):
            if value.shape == ():
                data[prefix + (key,)] = [value[()]]
            else:
                data[prefix + (key,)] = value[:]

    return data


def extract_dataset_from_h5(dataset) -> list:
    """
    Extracts values from a dataset within a h5 file

    Gets all values of a dataset in a h5 file and writes it to a list.

    :param dataset: dataset within a h5 file
    :return: list of all values in a dataset
    :rtype: list
    """
    data = [item.decode("utf-8") for item in dataset]

    return data


def add_values_to_summary(summary_path: Path, component_set: list = None):
    """
    Collect values of input cost parameters and relevant variables from HDF5 files and add them to the summary Excel file.

    Args:
        summary_path (Path or str): Path to the summary Excel file.
        component_set (list, optional): List of components to extract parameters and variables from.
            Defaults to ["Technologies", "Networks", "Import", "Export"].
    """

    if component_set is None:
        component_set = ["Technologies", "Networks", "Import", "Export"]

    summary_results = pd.read_excel(summary_path)

    # paths to results
    paths = {}
    for timestamp in summary_results["time_stamp"].unique():
        paths[timestamp] = list(
            summary_results.loc[
                summary_results["time_stamp"] == timestamp, "time_stamp"
            ].values
        )

    # dicts to store data
    output_dict = {}

    # Extract data from h5 files

    for case in paths:
        path = Path(case)
        hdf_file_path = path / "optimization_results.h5"
        output_dict[case] = {}
        if hdf_file_path.exists():
            with h5py.File(hdf_file_path, "r") as hdf_file:

                if "Technologies" in component_set:
                    data = extract_datasets_from_h5group(hdf_file["design/nodes"])
                    df = pd.DataFrame(data)
                    for period in df.columns.levels[0]:
                        for node in df.columns.levels[1]:
                            for tec in df.columns.levels[2]:
                                parameters = [
                                    "size",
                                    "capex_tot",
                                    "para_unitCAPEX",
                                    "para_fixCAPEX",
                                ]
                                for para in parameters:
                                    if (period, node, tec, para) in df.columns:
                                        tec_output = df[period, node, tec, para].iloc[0]
                                        output_name = f"{period}/{node}/{tec}/{para}"
                                        if output_name not in output_dict[case]:
                                            output_dict[case][output_name] = tec_output

                if "Networks" in component_set:
                    data = extract_datasets_from_h5group(hdf_file["design/networks"])
                    df = pd.DataFrame(data)
                    if not df.empty:
                        for period in df.columns.levels[0]:
                            for netw in df.columns.levels[1]:
                                for arc in df.columns.levels[2]:
                                    parameters = [
                                        "para_capex_gamma1",
                                        "para_capex_gamma2",
                                        "para_capex_gamma3",
                                        "para_capex_gamma4",
                                        "size",
                                        "capex",
                                    ]
                                    for para in parameters:
                                        output_name = f"{period}/{netw}/{arc}/{para}"
                                        arc_output = df[period, netw, arc, para].iloc[0]
                                        if output_name not in output_dict[case]:
                                            output_dict[case][output_name] = arc_output

                if "Import" in component_set:
                    data = extract_datasets_from_h5group(
                        hdf_file["operation/energy_balance"]
                    )
                    df = pd.DataFrame(data)
                    for period in df.columns.levels[0]:
                        for node in df[period].columns.levels[0]:
                            cars_at_node = (
                                df[period, node].columns.droplevel([1]).unique()
                            )
                            for car in cars_at_node:
                                parameters = ["import", "import_price"]
                                for para in parameters:
                                    car_output = df[period, node, car, para]
                                    if para == "import":
                                        car_output = sum(car_output)
                                        output_name = (
                                            f"{period}/{node}/{car}/{para}_tot"
                                        )
                                        if output_name not in output_dict[case]:
                                            output_dict[case][output_name] = car_output
                                    elif para == "import_price":
                                        car_output_mean = np.mean(car_output)
                                        car_output_std = np.std(car_output)
                                        output_name_mean = (
                                            f"{period}/{node}/{car}/{para}_mean"
                                        )
                                        output_name_std = (
                                            f"{period}/{node}/{car}/{para}_std"
                                        )
                                        if output_name_mean not in output_dict[case]:
                                            output_dict[case][
                                                output_name_mean
                                            ] = car_output_mean
                                        if output_name_std not in output_dict[case]:
                                            output_dict[case][
                                                output_name_std
                                            ] = car_output_std

                if "Export" in component_set:
                    data = extract_datasets_from_h5group(
                        hdf_file["operation/energy_balance"]
                    )
                    df = pd.DataFrame(data)
                    for period in df.columns.levels[0]:
                        for node in df[period].columns.levels[0]:
                            cars_at_node = (
                                df[period, node].columns.droplevel([1]).unique()
                            )
                            for car in cars_at_node:
                                parameters = ["export", "export_price"]
                                for para in parameters:
                                    car_output = df[period, node, car, para]
                                    if para == "export":
                                        car_output = sum(car_output)
                                        output_name = (
                                            f"{period}/{node}/{car}/{para}_tot"
                                        )
                                        if output_name not in output_dict[case]:
                                            output_dict[case][output_name] = car_output
                                    elif para == "export_price":
                                        car_output_mean = np.mean(car_output)
                                        car_output_std = np.std(car_output)
                                        output_name_mean = (
                                            f"{period}/{node}/{car}/{para}_mean"
                                        )
                                        output_name_std = (
                                            f"{period}/{node}/{car}/{para}_std"
                                        )
                                        if output_name_mean not in output_dict[case]:
                                            output_dict[case][
                                                output_name_mean
                                            ] = car_output_mean
                                        if output_name_std not in output_dict[case]:
                                            output_dict[case][
                                                output_name_std
                                            ] = car_output_std

    # Add new columns to summary_results
    output_df = pd.DataFrame(output_dict).T
    summary_results = summary_results.set_index("time_stamp")

    # Check for existing columns and overwrite them
    for col in output_df.columns:
        summary_results[col] = output_df[col]

    # Reset the index to ensure time_stamp is a column
    summary_results = summary_results.reset_index().rename(
        columns={"index": "folder_name"}
    )

    # Save the updated summary_results to the Excel file
    summary_results.to_excel(summary_path, index=False)



import pandas as pd
from pathlib import Path


def load_household_data():
    path = Path(__file__).parent
    path = path / "data/data_household.xlsx"
    return pd.read_excel(path, header=0, nrows=8760)


def load_network_city_data():
    path = Path(__file__).parent
    path = path / "data/data_network_city.xlsx"
    return pd.read_excel(path, header=0, nrows=8760)


def load_network_rural_data():
    path = Path(__file__).parent
    path = path / "data/data_network_rural.xlsx"
    return pd.read_excel(path, header=0, nrows=8760)



from .save_results import write_optimization_results_to_h5
from .read_results import (
    print_h5_tree,
    extract_dataset_from_h5,
    extract_datasets_from_h5group,
)
from .utilities import create_save_folder, create_unique_folder_name



from pyomo.environ import SolverFactory


def get_gurobi_parameters(solveroptions: dict):
    """
    Initiates the gurobi solver and defines solver parameters

    :param dict solveroptions: dict with solver parameters
    :return: Gurobi Solver
    """
    solver = SolverFactory(solveroptions["solver"]["value"], solver_io="python")
    solver.options["TimeLimit"] = solveroptions["timelim"]["value"] * 3600
    solver.options["MIPGap"] = solveroptions["mipgap"]["value"]
    solver.options["MIPFocus"] = solveroptions["mipfocus"]["value"]
    solver.options["Threads"] = solveroptions["threads"]["value"]
    solver.options["NodefileStart"] = solveroptions["nodefilestart"]["value"]
    solver.options["Method"] = solveroptions["method"]["value"]
    solver.options["Heuristics"] = solveroptions["heuristics"]["value"]
    solver.options["Presolve"] = solveroptions["presolve"]["value"]
    solver.options["BranchDir"] = solveroptions["branchdir"]["value"]
    solver.options["LPWarmStart"] = solveroptions["lpwarmstart"]["value"]
    solver.options["IntFeasTol"] = solveroptions["intfeastol"]["value"]
    solver.options["FeasibilityTol"] = solveroptions["feastol"]["value"]
    solver.options["Cuts"] = solveroptions["cuts"]["value"]
    solver.options["NumericFocus"] = solveroptions["numericfocus"]["value"]

    return solver


def get_glpk_parameters(solveroptions: dict):
    """
    Initiates the glpk solver and defines solver parameters

    :param dict solveroptions: dict with solver parameters
    :return: Gurobi Solver
    """
    solver = SolverFactory("glpk")

    return solver


def get_set_t(config: dict, model_block):
    """
    Returns the correct set_t for different clustering options

    :param dict config: config dict
    :param model_block: pyomo block holding set_t_full and set_t_clustered
    :return: set_t
    """
    if config["optimization"]["typicaldays"]["N"]["value"] == 0:
        return model_block.set_t_full
    elif config["optimization"]["typicaldays"]["method"]["value"] == 1:
        return model_block.set_t_clustered
    elif config["optimization"]["typicaldays"]["method"]["value"] == 2:
        return model_block.set_t_full


def get_hour_factors(config: dict, data, period: str) -> list:
    """
    Returns the correct hour factors to use for global balances

    :param dict config: config dict
    :param data: DataHandle
    :return: hour factors
    """
    if config["optimization"]["typicaldays"]["N"]["value"] == 0:
        return [1] * len(data.topology["time_index"]["full"])
    elif config["optimization"]["typicaldays"]["method"]["value"] == 1:
        return data.k_means_specs[period]["factors"]
    elif config["optimization"]["typicaldays"]["method"]["value"] == 2:
        return [1] * len(data.topology["time_index"]["full"])


def get_nr_timesteps_averaged(config: dict) -> int:
    """
    Returns the correct number of timesteps averaged

    :param dict config: config dict
    :return: nr_timesteps_averaged
    """
    if config["optimization"]["timestaging"]["value"] != 0:
        nr_timesteps_averaged = config["optimization"]["timestaging"]["value"]
    else:
        nr_timesteps_averaged = 1

    return nr_timesteps_averaged



import time
import pyomo.environ as pyo

import logging

log = logging.getLogger(__name__)


def annualize(r: float, t: int, year_fraction: float):
    """
    Calculates annualization factor

    :param float r: interest rate
    :param int t: lifetime
    :param flaot year_fraction: fraction of year modelled
    :return: annualization factor
    :rtype: float
    """
    if r == 0:
        annualization_factor = 1 / t
    else:
        annualization_factor = r / (1 - (1 / (1 + r) ** t))
    return annualization_factor * year_fraction


def set_discount_rate(config: dict, economics):
    """
    Sets the discount rate to either global or technology value

    :param dict config: dict containing model information
    :param economics: Economics class
    :return: CAPEX model
    :rtype: float
    """
    if not config["economic"]["global_discountrate"]["value"] == -1:
        discount_rate = config["economic"]["global_discountrate"]["value"]
    else:
        discount_rate = economics.discount_rate
    return discount_rate


def link_full_resolution_to_clustered(
    var_clustered, var_full, set_t_full, sequence, *other_sets
):
    """
    Links two variables (clustered and full)

    :param var_clustered: pyomo variable with clustered resolution
    :param var_full: pyomo variable with full resolution
    :param set_t_full: pyomo set containing timesteps
    :param sequence: order of typical days
    :param other_sets: other pyomo sets that variables are indexed by
    :return: pyomo constraint linking var_clustered and var_full
    """
    if not other_sets:

        def init_link_full_resolution(const, t):
            return var_full[t] == var_clustered[sequence[t - 1]]

        constraint = pyo.Constraint(set_t_full, rule=init_link_full_resolution)
    elif len(other_sets) == 1:
        set1 = other_sets[0]

        def init_link_full_resolution(const, t, set1):
            return var_full[t, set1] == var_clustered[sequence[t - 1], set1]

        constraint = pyo.Constraint(set_t_full, set1, rule=init_link_full_resolution)
    elif len(other_sets) == 2:
        set1 = other_sets[0]
        set2 = other_sets[1]

        def init_link_full_resolution(const, t, set1, set2):
            return var_full[t, set1, set2] == var_clustered[sequence[t - 1], set1, set2]

        constraint = pyo.Constraint(
            set_t_full, set1, set2, rule=init_link_full_resolution
        )

    return constraint


def perform_disjunct_relaxation(model_block, method: str = "gdp.bigm"):
    """
    Performs big-m transformation for respective component

    :param component: pyomo component
    :param str method: method to make transformation with.
    :return: component
    """
    log_msg = "\t\t\t" + method + " Transformation..."
    log.info(log_msg)
    start = time.time()
    xfrm = pyo.TransformationFactory(method)
    xfrm.apply_to(model_block)
    log_msg = (
        "\t\t\t"
        + method
        + " Transformation completed in "
        + str(round(time.time() - start))
        + " s"
    )

    log.info(log_msg)
    return model_block


def read_dict_value(dict: dict, key: str) -> str | int | float:
    """
    Reads a value from a dictonary or sets it to 1 if key is not in dict

    :param dict: dict
    :param key: dict key to check for
    :return:
    :rtype: str | int | float
    """
    dict_value = 1

    if dict:
        if key in dict:
            dict_value = dict[key]

    return dict_value


def determine_variable_scaling(model, model_block, f: dict, f_global):
    """
    Scale model block variables

    :param model: pyomo model
    :param model_block: pyomo model block
    :param dict f: individual scaling factors
    :param f_global: global scaling factors
    :return: model_block
    """
    for var in model_block.component_objects(pyo.Var, active=True):
        var_name = var.name.split(".")[-1]

        # check if var is integer
        var_is_integer = any([var[index].is_integer() for index in var.index_set()])

        if not var_is_integer:
            # Determine global scaling factor
            global_scaling_factor = f_global["energy_vars"]["value"] * read_dict_value(
                f, var_name
            )
            if "capex" in var_name or "opex" in var_name:
                global_scaling_factor = (
                    global_scaling_factor * f_global["cost_vars"]["value"]
                )
            model.scaling_factor[var] = global_scaling_factor

    return model


def determine_constraint_scaling(model, model_block, f: dict, f_global):
    """
    Scale model block constraints

    :param model: pyomo model
    :param model_block: pyomo model block
    :param dict f: individual scaling factors
    :param f_global: global scaling factors
    :return: model_block
    """
    for constr in model_block.component_objects(pyo.Constraint, active=True):
        const_name = constr.name.split(".")[-1]

        # Determine global scaling factor
        global_scaling_factor = (
            read_dict_value(f, const_name) * f_global["energy_vars"]["value"]
        )
        if "capex" in const_name or "opex" in const_name:
            global_scaling_factor = (
                global_scaling_factor * f_global["cost_vars"]["value"]
            )

        if not const_name.endswith("xor"):
            model.scaling_factor[constr] = global_scaling_factor

    return model


def get_attribute_from_dict(d: dict, key: str, value_other) -> str | float:
    """
    Takes an attribute from dict, if it doesnt exist replace with value_other

    :param dict d: dictonary
    :param str key: key to look for in dictonary
    :param value_other: if key is not in dict, return this value
    """
    if key in d:
        return d[key]
    else:
        return value_other



import json
import os
import pwlf
import numpy as np
from math import floor, log10
from statsmodels import api as sm
from pathlib import Path


def open_json(tec: str, load_path: str | Path) -> dict:
    """
    Reads technology data from json file

    :param str tec: Technology name
    :param str, Path load_path: directory to look for json in
    :return: dict with technology data from json
    :rtype: dict
    """
    # Read in JSON files
    for path, subdirs, files in os.walk(load_path):
        if "technology_data" in locals():
            break
        else:
            for name in files:
                if (tec + ".json") == name:
                    filepath = os.path.join(path, name)
                    with open(filepath) as json_file:
                        technology_data = json.load(json_file)
                    break

    # Assign name
    if "technology_data" in locals():
        technology_data["Name"] = tec
    else:
        raise Exception("There is no json data file for technology " + tec)

    return technology_data


def set_capex_model(config: dict, economics) -> int:
    """
    Sets the capex model of a technology

    Takes either the global capex model or the model defined in respective technology
    :param dict config: dict containing model information
    :param economics: Economics class
    :return: CAPEX model
    :rtype: int
    """
    capex_model = economics.capex_model
    if capex_model != 4:
        if config["economic"]["global_simple_capex_model"]["value"]:
            capex_model = 1

    return capex_model


def fit_linear_function(x: np.array, y: np.array) -> np.array:
    """
    Fits linear model to x and y data and returns coefficients

    :param np.array x: x data
    :param np.array y: y data
    :return: coefficients of OLS regression
    :rtype: np.array
    """
    linmodel = sm.OLS(y, x)
    linfit = linmodel.fit()
    coeff = linfit.params
    return coeff


def fit_piecewise_function(X: np.array, Y: np.array, nr_segments: int) -> dict:
    """
    Returns fitted parameters of a piecewise defined function with multiple y-series

    :param np.array X: x-values of data
    :param np.array Y: y-values of data
    :param int nr_seg: number of segments on piecewise defined function
    :return: x and y breakpoints, slope and intercept parameters of piecewise defined function
    :rtype: dict
    """

    def regress_piecewise(x, y, nr_segments, x_bp=None):
        """
        Returns fitted parameters of a piecewise defined function
        :param np.array X: x-values of data
        :param np.array y: y-values of data
        :param nr_seg: number of segments on piecewise defined function
        :return: x and y breakpoints, slope and intercept parameters of piecewise defined function
        """
        # Perform fit
        my_pwlf = pwlf.PiecewiseLinFit(x, y)
        if x_bp is None:
            my_pwlf.fit(nr_segments)
        else:
            my_pwlf.fit_with_breaks(x_bp)

        # retrieve data
        bp_x = my_pwlf.fit_breaks
        bp_y = my_pwlf.predict(bp_x)

        alpha1 = []
        alpha2 = []
        for seg in range(0, nr_segments):
            al1 = (bp_y[seg + 1] - bp_y[seg]) / (bp_x[seg + 1] - bp_x[seg])  # Slope
            al2 = (
                bp_y[seg]
                - (bp_y[seg + 1] - bp_y[seg]) / (bp_x[seg + 1] - bp_x[seg]) * bp_x[seg]
            )  # Intercept
            alpha1.append(al1)
            alpha2.append(al2)

        return bp_x, bp_y, alpha1, alpha2

    fit = {}

    for idx, car in enumerate(Y):
        fit[car] = {}
        y = np.array(Y[car])
        if idx == 0:
            bp_x, bp_y, alpha1, alpha2 = regress_piecewise(X, y, nr_segments)
            bp_x0 = bp_x
        else:
            bp_x, bp_y, alpha1, alpha2 = regress_piecewise(X, y, nr_segments, bp_x0)

        fit[car]["alpha1"] = [sig_figs(float(num), 4) for num in alpha1]
        fit[car]["alpha2"] = [sig_figs(float(num), 4) for num in alpha2]
        fit[car]["bp_y"] = [sig_figs(float(num), 4) for num in bp_y]
        fit[car]["bp_x"] = [sig_figs(float(num), 4) for num in bp_x]

    return fit


def sig_figs(x: float, precision: int):
    """
    Rounds a number to number of significant figures

    :param float x: number to round
    :param int precision: rounding precision
    :return: rounded number
    :rtype: float
    """

    x = float(x)
    precision = int(precision)

    if x == 0:
        rounded = 0
    else:
        rounded = round(x, -int(floor(log10(abs(x)))) + (precision - 1))

    return rounded



import warnings
import pandas as pd
import numpy as np
import shutil
import os
import json
import requests
from timezonefinder import TimezoneFinder
from pathlib import Path


def load_climate_data_from_api(folder_path: str | Path, dataset: str = "JRC"):
    """
    Reads in climate data for a full year from a folder containing node data and writes it to the respective file.

    Reads in climate data for a full year from a folder containing node data,
    where each node data is stored in a subfolder and node locations are provided in a CSV file named NodeLocations.csv.
    The data is written to the file

    :param str folder_path: Path to the folder containing node data and NodeLocations.csv
    :param str dataset: Dataset to import from, can be JRC (only onshore)
    """
    # Convert to Path
    if isinstance(folder_path, str):
        folder_path = Path(folder_path)

    # Read NodeLocations.csv with node column as index
    node_locations_path = os.path.join(folder_path, "NodeLocations.csv")
    node_locations_df = pd.read_csv(
        node_locations_path, sep=";", names=["node", "lon", "lat", "alt"], header=0
    )

    if node_locations_df.isnull().values.any():
        raise Exception("Please specify longitude, latitude and altitude for each node")

    # Read nodes and investment_periods from the JSON file
    json_file_path = os.path.join(folder_path, "Topology.json")
    with open(json_file_path, "r") as json_file:
        topology = json.load(json_file)

    year = (
        int(topology["start_date"].split("-")[0])
        if topology["start_date"]
        else "typical_year"
    )

    for period in topology["investment_periods"]:
        for node_name in topology["nodes"]:
            # Read lon, lat, and alt for this node name from node_locations_df
            node_data = node_locations_df[node_locations_df["node"] == node_name]
            lon = node_data["lon"].values[0]
            lat = node_data["lat"].values[0]
            alt = node_data["alt"].values[0]

            if dataset == "JRC":
                # Fetch climate data for the node
                data = import_jrc_climate_data(lon, lat, year, alt)
            else:
                raise Exception("Other APIs are not available")

            # Write data to CSV file
            output_folder = os.path.join(folder_path, period, "node_data", node_name)
            output_file = os.path.join(output_folder, "ClimateData.csv")
            existing_data = pd.read_csv(output_file, sep=";")

            # Fill in existing data with data from the fetched DataFrame based on column names
            for column, value in data["dataframe"].items():
                existing_data[column] = value.values[: len(existing_data)]

            # Save the updated data back to ClimateData.csv
            existing_data.to_csv(output_file, index=False, sep=";")


def fill_carrier_data(
    folder_path: str | Path,
    value_or_data: float | pd.DataFrame,
    columns: list = [],
    carriers: list = [],
    nodes: list = [],
    investment_periods: list = None,
):
    """
    Updates carrier data for a time series based on a provided value or DataFrame and writes it to file.

    Allows you to update Demand, Import limit, Export limit, Import price,
    Export price, Import emission factor, Export emission factor and/or Generic production.

    :param str folder_path: Path to the folder containing the case study data
    :param float | pd.DataFrame value_or_data: A float value to be applied or a DataFrame containing the new values for the carrier data
    :param list columns: Name of the columns that need to be changed
    :param list investment_periods: Name of investment periods to be changed
    :param list nodes: Name of the nodes that need to be changed
    :param list carriers: Name of the carriers that need to be changed
    """
    # Convert to Path
    if isinstance(folder_path, str):
        folder_path = Path(folder_path)

    # Read the topology json file
    json_file_path = folder_path / "Topology.json"
    with open(json_file_path, "r") as json_file:
        topology = json.load(json_file)

    # Define options
    column_options = [
        "Demand",
        "Import limit",
        "Export limit",
        "Import price",
        "Export price",
        "Import emission factor",
        "Export emission factor",
        "Generic production",
    ]

    for period in (
        investment_periods if investment_periods else topology["investment_periods"]
    ):
        for node_name in nodes if nodes else topology["nodes"]:
            for car in carriers if carriers else topology["carriers"]:

                # Write data to CSV file
                output_folder = (
                    folder_path / period / "node_data" / node_name / "carrier_data"
                )
                filename = car + ".csv"
                output_file = output_folder / filename
                existing_data = pd.read_csv(output_file, sep=";")

                # Fill in existing data with either a constant value or data from the provided DataFrame
                for column in columns if columns else column_options:
                    if isinstance(value_or_data, pd.DataFrame):
                        if column in value_or_data.columns:
                            existing_data[column] = value_or_data[column].values
                        else:
                            raise ValueError(
                                f"Column {column} not found in the provided DataFrame"
                            )
                    else:
                        existing_data[column] = value_or_data * np.ones(
                            len(existing_data)
                        )

                # Save the updated data back to the CSV file
                output_file.parent.mkdir(
                    parents=True, exist_ok=True
                )  # Ensure directory exists
                existing_data.to_csv(output_file, index=False, sep=";")


def copy_technology_data(folder_path: str | Path, tec_data_path: str | Path = None):
    """
    Copies technology JSON files to the node folder for each node and investment period.

    This function reads the topology JSON file to determine the existing and new technologies at each node for
    each investment period. It then searches for the corresponding JSON files in the specified `tec_data_path`
    folder (and its subfolders) using the technology names and copies them to the output folder.

    :param str | Path folder_path: Path to the folder containing the case study data.
    :param str | Path tec_data_path: Path to the folder containing the technology data.
    """
    # Convert to Path
    if isinstance(folder_path, str):
        folder_path = Path(folder_path)

    if tec_data_path is None:
        tec_data_path = Path(
            os.path.join(os.path.dirname(__file__) + "/../data/technology_data")
        )
    else:
        if isinstance(tec_data_path, str):
            tec_data_path = Path(tec_data_path)

    # Reads the topology JSON file
    json_file_path = folder_path / "Topology.json"
    with open(json_file_path, "r") as json_file:
        topology = json.load(json_file)

    for period in topology["investment_periods"]:
        for node_name in topology["nodes"]:
            # Read the JSON technology file
            json_tec_file_path = (
                folder_path / period / "node_data" / node_name / "Technologies.json"
            )
            with open(json_tec_file_path, "r") as json_tec_file:
                json_tec = json.load(json_tec_file)
            tecs_at_node = list(json_tec["existing"].keys()) + json_tec["new"]

            output_folder = (
                folder_path / period / "node_data" / node_name / "technology_data"
            )
            # Copy JSON files corresponding to technology names to output folder
            for tec_name in tecs_at_node:
                _copy_data(tec_data_path, tec_name, output_folder)


def copy_network_data(folder_path: str | Path, ntw_data_path: str | Path = None):
    """
    Copies network JSON files to the network_data folder for each investment period.

    This function reads the topology JSON file to determine the existing and new networks for
    each investment period. It then searches for the corresponding JSON files in the specified `ntw_data_path`
    folder (and its subfolders) using the network names and copies them to folder_path.

    :param str | Path folder_path: Path to the folder containing the case study data.
    :param str | Path ntw_data_path: Path to the folder containing the network data (if left
    empty, standard folder is used).
    :return: None
    """
    # Convert to Path
    if isinstance(folder_path, str):
        folder_path = Path(folder_path)

    if ntw_data_path is None:
        ntw_data_path = Path(
            os.path.join(os.path.dirname(__file__) + "/../data/network_data")
        )
    else:
        if isinstance(ntw_data_path, str):
            ntw_data_path = Path(ntw_data_path)

    # Reads the topology JSON file
    json_file_path = folder_path / "Topology.json"
    with open(json_file_path, "r") as json_file:
        topology = json.load(json_file)

    for period in topology["investment_periods"]:
        # Read the JSON network file
        json_ntw_file_path = folder_path / period / "Networks.json"
        with open(json_ntw_file_path, "r") as json_ntw_file:
            json_ntw = json.load(json_ntw_file)
        ntws_at_node = json_ntw["existing"] + json_ntw["new"]

        output_folder = folder_path / period / "network_data"
        # Copy JSON files corresponding to technology names to output folder
        for ntw_name in ntws_at_node:
            _copy_data(ntw_data_path, ntw_name, output_folder)


def find_json_path(data_path: str | Path, name: str) -> Path | None:
    """
    Search for a JSON file with the given technology name in the specified path and its subfolders.

    :param str data_path: Path to the folder containing technology JSON files.
    :param str name: Name of the technology.
    :return: Path to the JSON file if found, otherwise None.
    """
    for root, dirs, files in os.walk(data_path.resolve()):
        for file in files:
            if file.lower() == f"{name.lower()}.json":
                return Path(root) / Path(file)


def import_jrc_climate_data(
    lon: float, lat: float, year: int | str, alt: float
) -> dict:
    """
    Reads in climate data for a full year from `JRC PVGIS <https://re.jrc.ec.europa.eu/pvg_tools/en/>`_.

    The returned dataframe is consistent with the modelhub format requirements.

    :param float lon: longitude of node - the api will read data for this location
    :param float lat: latitude of node - the api will read data for this location
    :param int year: optional, needs to be in range of data available. If nothing is specified, a typical year \
    will be loaded
    :param float alt: altitude of location specified
    :return: dict containing information on the location (altitude, longitude, latitude and a dataframe \
    containing climate data (ghi = global horizontal irradiance, dni = direct normal irradiance, \
    dhi = diffuse horizontal irradiance, rh = relative humidity, temp_air = air temperature, ws = wind speed at \
    specified hight. Wind speed is returned as a dict for different heights.
    :rtype: dict
    """
    # get time zone
    tf = TimezoneFinder()

    # Specify year import, lon, lat
    if year == "typical_year":
        parameters = {"lon": lon, "lat": lat, "outputformat": "json"}
        time_index = pd.date_range(start="2001-01-01 00:00", freq="1h", periods=8760)
    else:
        parameters = {"lon": lon, "lat": lat, "year": year, "outputformat": "json"}
        time_index = pd.date_range(
            start=str(year) + "-01-01 00:00", end=str(year) + "-12-31 23:00", freq="1h"
        )

    # Get data from JRC dataset
    answer = dict()
    print("Importing Climate Data...")
    response = requests.get("https://re.jrc.ec.europa.eu/api/tmy?", params=parameters)
    if response.status_code == 200:
        print("Importing Climate Data successful")
    else:
        raise Exception(response)
    data = response.json()
    climate_data = data["outputs"]["tmy_hourly"]

    # Compile return dict
    answer["longitude"] = lon
    answer["latitude"] = lat
    answer["altitude"] = alt

    ghi = []
    dni = []
    dhi = []
    rh = []
    temp_air = []
    wind_speed = dict()
    wind_speed["10"] = []

    for t_interval in climate_data:
        ghi.append(t_interval["G(h)"])
        dni.append(t_interval["Gb(n)"])
        dhi.append(t_interval["Gd(h)"])
        rh.append(t_interval["RH"])
        temp_air.append(t_interval["T2m"])
        wind_speed["10"].append(t_interval["WS10m"])

    answer["dataframe"] = pd.DataFrame(
        np.array([ghi, dni, dhi, temp_air, rh]).T,
        columns=["ghi", "dni", "dhi", "temp_air", "rh"],
        index=time_index,
    )
    for ws in wind_speed:
        answer["dataframe"]["ws" + str(ws)] = wind_speed[ws]

    return answer


def _copy_data(path, json_name, output_folder):
    """
    Finds the JSON files and copies it to the desired output folder.

    This function finds the JSON file of the component, and it copies it to the output folder.
    It also reads the JSON file, checks if CCS is possible and, if so, it copies the
    corresponding CCS JSON file to the output folder.

    :param str | Path path: Path to the folder containing the case study data.
    :param str | Component name json_name: Name of the JSON file.
    :param str | Path output_folder: Path to the folder containing the technology data.
    """
    component_json_path = find_json_path(path, json_name)
    if component_json_path:
        shutil.copy(component_json_path, output_folder)
        # Adding copy CCS when possible
        with open(component_json_path, "r") as json_data_file:
            json_data = json.load(json_data_file)
        if (
            "ccs" in json_data["Performance"]
            and json_data["Performance"]["ccs"]["possible"]
        ):
            ccs_name = json_data["Performance"]["ccs"]["ccs_type"]
            ccs_json_path = find_json_path(path, ccs_name)
            ccs_output_path = os.path.join(
                output_folder, os.path.basename(ccs_json_path)
            )
            # Copy CCS JSON if it doesn't exist yet
            if not os.path.exists(ccs_output_path):
                _copy_data(path, ccs_name, output_folder)
    else:
        warnings.warn(f"{json_name} not found")



import pyomo.core
import pyomo.environ as pyo
import pyomo.gdp as gdp
import numpy as np
import pandas as pd

from ..technology import Technology, set_capex_model
from ....components.utilities import (
    annualize,
    set_discount_rate,
    get_attribute_from_dict,
    link_full_resolution_to_clustered,
)


class Stor(Technology):
    """
    Storage Technology

    This model resembles a storage technology.
    Note that this technology only works for one carrier, and thus the carrier index is dropped in the below notation.

    **Variable declarations:**

    - ``var_storage_level``: Storage level in :math:`t`: :math:`E_t`

    - ``var_capacity_charge``: Charging capacity

    - ``var_capacity_discharge``: Discharging capacity

    **Constraint declarations:**

    The following constants are used:

    - :math:`{\\eta}_{in}`: Charging efficiency

    - :math:`{\\eta}_{out}`: Discharging efficiency

    - :math:`{\\lambda_1}`: Self-Discharging coefficient (independent of environment)

    - :math:`{\\lambda_2(\\Theta)}`: Self-Discharging coefficient (dependent on environment)

    - :math:`Input_{max}`: Maximal charging capacity

    - :math:`Output_{max}`: Maximal discharging capacity

    - Size constraint:

      .. math::
        E_{t} \\leq S

    - Maximal charging and discharging:

      .. math::
        Input_{t} \\leq Input_{max}

      .. math::
        Output_{t} \\leq Output_{max}

    - Storage level calculation:

      .. math::
        E_{t} = E_{t-1} * (1 - \\lambda_1) - \\lambda_2(\\Theta) * E_{t-1} + {\\eta}_{in} * Input_{t} - 1 / {\\eta}_{out} * Output_{t}

    - If ``allow_only_one_direction == 1``, simultaneous charging and discharging is limited
      using Equation 24 from Morales-España, Germán & Hernandez, Ricardo & Helistö, Niina & Kiviluoma,
      Juha. (2022). LP Formulation for Optimal Investment and Operation of Storage Including Reserves.
      10.13140/RG.2.2.27048.03840.

    - If ``allow_only_one_direction_precise == 1``, then only input or output can be unequal
      to zero in each respective time step (otherwise, simultaneous charging and
      discharging can lead to unwanted 'waste' of energy/material).

     - If in ``Flexibility`` the ``power_energy_ratio == fixed``, then the capacity of
       the charging and discharging power is fixed as a ratio of the energy capacity.
       Thus:

       .. math::
         Input_{max} = \\gamma_{charging} * S

    - If in 'Flexibility' the "power_energy_ratio == flexratio", then the
      capacity of the charging and discharging power is a variable in the
      optimization. In this case, the charging and discharging rates specified in the
      json file are the maximum installed capacities as a ratio of the energy
      capacity. The model will optimize the charging and discharging capacities,
      based on the incorporation of these components in the CAPEX function.

    - If in 'Flexibility' the "power_energy_ratio == fixedratio", then the
      capacity of the charging and discharging power is a fraction of the installed capacity.
      In this case, the charging and discharging rates specified in the
      json file are a ratio of the energy capacity.

    - If in 'Flexibility' the "power_energy_ratio == fixedcapacity", then the
      capacity of the charging and discharging power is a fixed input parameter.
      In this case, the charging and discharging rates specified in the
      json file in the same unit as the input and output.

    - If an energy consumption for charging or dis-charging process is given, the respective carrier input is:

      .. math::
        Input_{t, car} = cons_{car, in} Input_{t}

      .. math::
        Input_{t, car} = cons_{car, out} Output_{t}

    - CAPEX is given by three contributions, the CAPEX of the charging capacity, the CAPEX of the discharging capacity,
      and the CAPEX of the energy capacity (i.e., the size of the storage).

      .. math::
        CAPEX_{chargeCapacity} = chargeCapacity * unitCost_{chargeCapacity}

      .. math::
        CAPEX_{dischargeCapacity} = dischargeCapacity * unitCost_{dischargeCapacity}

      .. math::
        CAPEX_{storSize} = storSize * unitCost_{storSize}

    - Additionally, ramping rates of the technology can be constraint (for input and
      output).

      .. math::
         -rampingrate \\leq Input_{t, maincar} - Input_{t-1, maincar} \\leq rampingrate

      .. math::
         -rampingrate \\leq \\sum(Output_{t, car}) - \\sum(Output_{t-1, car}) \\leq
         rampingrate


    """

    def __init__(self, tec_data: dict):
        """
        Constructor

        :param dict tec_data: technology data
        """
        super().__init__(tec_data)

        self.component_options.emissions_based_on = "input"
        self.component_options.main_input_carrier = tec_data["Performance"][
            "main_input_carrier"
        ]

        self.flexibility_data = tec_data["Flexibility"]

    def fit_technology_performance(self, climate_data: pd.DataFrame, location: dict):
        """
        Fits conversion technology type STOR and fills in the fitted parameters in a dict

        :param pd.Dataframe climate_data: dataframe containing climate data
        :param dict location: dict containing location details
        """
        super(Stor, self).fit_technology_performance(climate_data, location)

        # For a flexibly optimized storage technology (i.e., not a fixed P-E ratio), an adapted CAPEX function is used
        # to account for charging and discharging capacity costs.
        if self.flexibility_data["power_energy_ratio"] == "flexratio":
            self.economics.capex_model = 4
        if self.flexibility_data["power_energy_ratio"] not in [
            "flexratio",
            "fixedratio",
            "fixedcapacity",
        ]:
            raise Warning(
                "power_energy_ratio should be either flexible ('flexratio') or fixed ('fixedratio') or as capacity ('fixedcapacity')"
            )

        # Coefficients
        theta = self.input_parameters.performance_data["performance"]["theta"]
        ambient_loss_factor = (65 - climate_data["temp_air"]) / (90 - 65) * theta

        self.processed_coeff.time_dependent_full["ambient_loss_factor"] = (
            ambient_loss_factor.to_numpy()
        )

        for par in self.input_parameters.performance_data["performance"]:
            if not par == "theta":
                self.processed_coeff.time_independent[par] = (
                    self.input_parameters.performance_data["performance"][par]
                )

        self.processed_coeff.time_independent["charge_rate"] = self.flexibility_data[
            "charge_rate"
        ]
        self.processed_coeff.time_independent["discharge_rate"] = self.flexibility_data[
            "discharge_rate"
        ]
        if (
            "energy_consumption"
            in self.input_parameters.performance_data["performance"]
        ):
            self.processed_coeff.time_independent["energy_consumption"] = (
                self.input_parameters.performance_data["performance"][
                    "energy_consumption"
                ]
            )

    def _calculate_bounds(self):
        """
        Calculates the bounds of the variables used
        """
        super(Stor, self)._calculate_bounds()

        time_steps = len(self.set_t_performance)

        # Output Bounds
        for car in self.component_options.output_carrier:
            self.bounds["output"][car] = np.column_stack(
                (
                    np.zeros(shape=(time_steps)),
                    np.ones(shape=(time_steps))
                    * self.flexibility_data["discharge_rate"],
                )
            )
        # Input Bounds
        for car in self.component_options.input_carrier:
            if car == self.component_options.main_input_carrier:
                self.bounds["input"][car] = np.column_stack(
                    (
                        np.zeros(shape=(time_steps)),
                        np.ones(shape=(time_steps))
                        * self.flexibility_data["charge_rate"],
                    )
                )
            else:
                if (
                    "energy_consumption"
                    in self.input_parameters.performance_data["performance"]
                ):
                    energy_consumption = self.input_parameters.performance_data[
                        "performance"
                    ]["energy_consumption"]
                    self.bounds["input"][car] = np.column_stack(
                        (
                            np.zeros(shape=(time_steps)),
                            np.ones(shape=(time_steps))
                            * self.flexibility_data["charge_rate"]
                            * energy_consumption["in"][car],
                        )
                    )

    def construct_tech_model(self, b_tec, data: dict, set_t_full, set_t_clustered):
        """
        Adds constraints to technology blocks for tec_type STOR, resembling a storage technology

        :param b_tec: pyomo block with technology model
        :param dict data: data containing model configuration
        :param set_t_full: pyomo set containing timesteps
        :param set_t_clustered: pyomo set containing clustered timesteps
        :return: pyomo block with technology model
        """

        super(Stor, self).construct_tech_model(b_tec, data, set_t_full, set_t_clustered)

        config = data["config"]

        # DATA OF TECHNOLOGY
        coeff_td = self.processed_coeff.time_dependent_used
        coeff_ti = self.processed_coeff.time_independent
        dynamics = self.processed_coeff.dynamics

        # sequence_storage = self.sequence
        if config["optimization"]["typicaldays"]["N"]["value"] == 0:
            sequence_storage = self.sequence
        elif config["optimization"]["typicaldays"]["method"]["value"] == 1:
            sequence_storage = data["k_means_specs"]["sequence"]
        elif config["optimization"]["typicaldays"]["method"]["value"] == 2:
            sequence_storage = self.sequence

        if config["optimization"]["timestaging"]["value"] != 0:
            nr_timesteps_averaged = config["optimization"]["timestaging"]["value"]
        else:
            nr_timesteps_averaged = 1

        # Additional parameters
        eta_in = coeff_ti["eta_in"]
        eta_out = coeff_ti["eta_out"]
        eta_lambda = coeff_ti["lambda"]
        charge_rate = coeff_ti["charge_rate"]
        discharge_rate = coeff_ti["discharge_rate"]
        ambient_loss_factor = coeff_td["ambient_loss_factor"]

        # Additional decision variables
        b_tec.var_storage_level = pyo.Var(
            self.set_t_full,
            domain=pyo.NonNegativeReals,
            bounds=(b_tec.para_size_min, b_tec.para_size_max),
        )

        # Size constraint
        def init_size_constraint(const, t):
            return b_tec.var_storage_level[t] <= b_tec.var_size

        b_tec.const_size = pyo.Constraint(self.set_t_full, rule=init_size_constraint)

        # Storage level calculation
        def init_storage_level(const, t):
            if t == 1:
                # couple first and last time interval: storageLevel[1] ==
                # storageLevel[end] * (1-self_discharge)^nr_timesteps_averaged +
                # - storageLevel[end] * ambient_loss_factor[end-1]^nr_timesteps_averaged +
                # + (eta_in * input[] +1/eta_out * output[]) *
                # (sum (1-self_discharge)^i for i in [0, nr_timesteps_averaged])
                #
                # soc[1] = soc[end] + input[seq] - output[seq]

                return b_tec.var_storage_level[t] == b_tec.var_storage_level[
                    max(self.set_t_full)
                ] * (1 - eta_lambda) ** nr_timesteps_averaged - b_tec.var_storage_level[
                    max(self.set_t_full)
                ] * ambient_loss_factor[
                    sequence_storage[t - 1] - 1
                ] ** nr_timesteps_averaged + (
                    eta_in
                    * self.input[
                        sequence_storage[t - 1],
                        self.component_options.main_input_carrier,
                    ]
                    - 1
                    / eta_out
                    * self.output[
                        sequence_storage[t - 1],
                        self.component_options.main_input_carrier,
                    ]
                ) * sum(
                    (1 - eta_lambda) ** i for i in range(0, nr_timesteps_averaged)
                )
            else:  # all other time intervals
                return b_tec.var_storage_level[t] == b_tec.var_storage_level[t - 1] * (
                    1 - eta_lambda
                ) ** nr_timesteps_averaged - b_tec.var_storage_level[
                    t
                ] * ambient_loss_factor[
                    sequence_storage[t - 1] - 1
                ] ** nr_timesteps_averaged + (
                    eta_in
                    * self.input[
                        sequence_storage[t - 1],
                        self.component_options.main_input_carrier,
                    ]
                    - 1
                    / eta_out
                    * self.output[
                        sequence_storage[t - 1],
                        self.component_options.main_input_carrier,
                    ]
                ) * sum(
                    (1 - eta_lambda) ** i for i in range(0, nr_timesteps_averaged)
                )

        b_tec.const_storage_level = pyo.Constraint(
            self.set_t_full, rule=init_storage_level
        )

        # CONSTRAINTS FOR BIDIRECTIONAL STORAGE
        if self.component_options.allow_only_one_direction:

            # Cut according to Morales-Espana "LP Formulation for Optimal Investment and
            # Operation of Storage Including Reserves"
            def init_cut_bidirectional(const, t):
                # output[t]/discharge_rate + input[t]/charge_rate <= storSize
                return (
                    self.output[t, self.component_options.main_input_carrier]
                    / discharge_rate
                    + self.input[t, self.component_options.main_input_carrier]
                    / charge_rate
                    <= b_tec.var_size
                )

            # Changes to cut from Morales-Espana for fixed charging and discharging capacities
            def init_cut_bidirectional_fix1(const, t):
                return (
                    self.output[t, self.component_options.main_input_carrier]
                    + self.input[t, self.component_options.main_input_carrier]
                    <= b_tec.var_size
                )

            def init_cut_bidirectional_fix2(const, t):
                return (
                    self.output[t, self.component_options.main_input_carrier]
                    + self.input[t, self.component_options.main_input_carrier]
                    <= charge_rate + discharge_rate
                )

            if self.flexibility_data["power_energy_ratio"] == "fixedcapacity":
                b_tec.const_cut_bidirectional1 = pyo.Constraint(
                    self.set_t_performance, rule=init_cut_bidirectional_fix1
                )
                b_tec.const_cut_bidirectional2 = pyo.Constraint(
                    self.set_t_performance, rule=init_cut_bidirectional_fix2
                )
            else:
                b_tec.const_cut_bidirectional = pyo.Constraint(
                    self.set_t_performance, rule=init_cut_bidirectional
                )

            if self.component_options.allow_only_one_direction_precise:

                self.big_m_transformation_required = 1
                s_indicators = range(0, 2)

                def init_input_output(dis, t, ind):
                    if ind == 0:  # input only

                        def init_output_to_zero(const, car_output):
                            return self.output[t, car_output] == 0

                        dis.const_output_to_zero = pyo.Constraint(
                            b_tec.set_output_carriers, rule=init_output_to_zero
                        )

                    elif ind == 1:  # output only

                        def init_input_to_zero(const, car_input):
                            return self.input[t, car_input] == 0

                        dis.const_input_to_zero = pyo.Constraint(
                            b_tec.set_input_carriers, rule=init_input_to_zero
                        )

                b_tec.dis_input_output = gdp.Disjunct(
                    self.set_t_performance, s_indicators, rule=init_input_output
                )

                # Bind disjuncts
                def bind_disjunctions(dis, t):
                    return [b_tec.dis_input_output[t, i] for i in s_indicators]

                b_tec.disjunction_input_output = gdp.Disjunction(
                    self.set_t_performance, rule=bind_disjunctions
                )

        # Maximal charging and discharging rates
        def init_maximal_charge(const, t):
            if self.flexibility_data["power_energy_ratio"] == "fixedcapacity":
                return (
                    self.input[t, self.component_options.main_input_carrier]
                    <= charge_rate
                )
            else:
                return (
                    self.input[t, self.component_options.main_input_carrier]
                    <= b_tec.var_capacity_charge
                )

        b_tec.const_max_charge = pyo.Constraint(
            self.set_t_performance, rule=init_maximal_charge
        )

        def init_maximal_discharge(const, t):
            if self.flexibility_data["power_energy_ratio"] == "fixedcapacity":
                return (
                    self.output[t, self.component_options.main_input_carrier]
                    <= discharge_rate
                )
            else:
                return (
                    self.output[t, self.component_options.main_input_carrier]
                    <= b_tec.var_capacity_discharge
                )

        b_tec.const_max_discharge = pyo.Constraint(
            self.set_t_performance, rule=init_maximal_discharge
        )

        # if the charging / discharging rates are fixed or flexible as a ratio of the energy capacity:
        if not self.flexibility_data["power_energy_ratio"] == "fixedcapacity":

            def init_max_capacity_charge(const):
                if self.flexibility_data["power_energy_ratio"] == "fixedratio":
                    return b_tec.var_capacity_charge == charge_rate * b_tec.var_size
                else:
                    return b_tec.var_capacity_charge <= charge_rate * b_tec.var_size

            b_tec.const_max_cap_charge = pyo.Constraint(rule=init_max_capacity_charge)

            def init_max_capacity_discharge(const):
                if self.flexibility_data["power_energy_ratio"] == "fixedratio":
                    # dischargeCapacity == dischargeRate * storSize
                    return (
                        b_tec.var_capacity_discharge == discharge_rate * b_tec.var_size
                    )
                else:
                    # dischargeCapacity <= dischargeRate * storSize
                    return (
                        b_tec.var_capacity_discharge <= discharge_rate * b_tec.var_size
                    )

            b_tec.const_max_cap_discharge = pyo.Constraint(
                rule=init_max_capacity_discharge
            )

        # Energy consumption charging/discharging
        if "energy_consumption" in coeff_ti:
            energy_consumption = coeff_ti["energy_consumption"]
            if "in" in energy_consumption:
                b_tec.set_energyconsumption_carriers_in = pyo.Set(
                    initialize=list(energy_consumption["in"].keys())
                )

                def init_energyconsumption_in(const, t, car):
                    # e.g electricity_cons[t] == input[t] * energy_cons[electricity]
                    return (
                        self.input[t, car]
                        == self.input[t, self.component_options.main_input_carrier]
                        * energy_consumption["in"][car]
                    )

                b_tec.const_energyconsumption_in = pyo.Constraint(
                    self.set_t_performance,
                    b_tec.set_energyconsumption_carriers_in,
                    rule=init_energyconsumption_in,
                )

            if "out" in energy_consumption:
                b_tec.set_energyconsumption_carriers_out = pyo.Set(
                    initialize=energy_consumption["out"].keys()
                )

                def init_energyconsumption_out(const, t, car):
                    # NOTE: even though we call it "energy consumption", this constraint actually
                    # simulates for example the production of electricity from a salt cavern H2 storage,
                    # where H2 is stored at e.g. 130bar and released at e.g. 40bar; the pressure difference can be
                    # used to drive a trubine and create electricity.
                    # e.g electricity_prod[t] == output[t] * energy_cons[electricity]
                    return (
                        self.output[t, car]
                        == self.output[t, self.component_options.main_input_carrier]
                        * energy_consumption["out"][car]
                    )

                b_tec.const_energyconsumption_out = pyo.Constraint(
                    self.set_t_performance,
                    b_tec.set_energyconsumption_carriers_out,
                    rule=init_energyconsumption_out,
                )

        # RAMPING RATES
        if "ramping_time" in dynamics:
            if not dynamics["ramping_time"] == -1:
                b_tec = self._define_ramping_rates(b_tec, data, sequence_storage)

        return b_tec

    def _define_size(self, b_tec):
        """
        Adds variables related to technology size specific to storage technologies (i.e., variables for the charging and
        discharging capacities). Regular technology parameters and variables are taken from the parent class (Technology).

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        super(Stor, self)._define_size(b_tec)

        coeff_ti = self.processed_coeff.time_independent
        charge_rate = coeff_ti["charge_rate"]
        discharge_rate = coeff_ti["discharge_rate"]

        if not self.flexibility_data["power_energy_ratio"] == "fixedcapacity":
            b_tec.var_capacity_charge = pyo.Var(
                domain=pyo.NonNegativeReals,
                bounds=(0, b_tec.para_size_max * charge_rate),
            )
            b_tec.var_capacity_discharge = pyo.Var(
                domain=pyo.NonNegativeReals,
                bounds=(0, b_tec.para_size_max * discharge_rate),
            )

        return b_tec

    def _define_capex_parameters(self, b_tec, data: dict):
        """
        Defines the capex parameters for a flexible storage technology of capex model 4 (for all other capex models,
        the parameters are defined in the Technology class).

        For capex model 4:
        - para_unit_capex_charging_cap: the capex per unit of charging capacity (i.e., per MW charging).
        - para_unit_capex_charging_cap_annual: annualized capex per unit of charging capacity.
        - para_unit_capex_discharging_cap: the capex per unit of discharging capacity (i.e., per MW discharging).
        - para_unit_capex_discharging_cap_annual: annualized capex per unit of discharging capacity.
        - para_unit_capex: capex per unit of energy storage capacity (i.e., per MWh storage).
        - para_unit_capex_annual: annualized capex per unit of storage capacity.

        :param b_tec: pyomo block with technology model
        :param dict data: dict containing model information
        :return: pyomo block with technology model
        """
        super(Stor, self)._define_capex_parameters(b_tec, data)

        config = data["config"]
        economics = self.economics

        # CAPEX PARAMETERS
        capex_model = set_capex_model(config, economics)

        if capex_model == 4:
            discount_rate = set_discount_rate(config, economics)
            fraction_of_year_modelled = data["topology"]["fraction_of_year_modelled"]
            annualization_factor = annualize(
                discount_rate, economics.lifetime, fraction_of_year_modelled
            )
            flexibility = self.flexibility_data

            # additional parameters needed for a flexible storage system
            b_tec.para_unit_capex_charging_cap = pyo.Param(
                domain=pyo.Reals,
                initialize=flexibility["capex_charging_power"],
                mutable=True,
            )
            b_tec.para_unit_capex_discharging_cap = pyo.Param(
                domain=pyo.Reals,
                initialize=flexibility["capex_discharging_power"],
                mutable=True,
            )
            b_tec.para_unit_capex = pyo.Param(
                domain=pyo.Reals,
                initialize=economics.capex_data["unit_capex"],
                mutable=True,
            )

            # annualized parameters for a flexible storage system
            b_tec.para_unit_capex_charging_cap_annual = pyo.Param(
                domain=pyo.Reals,
                initialize=(annualization_factor * b_tec.para_unit_capex_charging_cap),
                mutable=True,
            )
            b_tec.para_unit_capex_discharging_cap_annual = pyo.Param(
                domain=pyo.Reals,
                initialize=(
                    annualization_factor * b_tec.para_unit_capex_discharging_cap
                ),
                mutable=True,
            )
            b_tec.para_unit_capex_annual = pyo.Param(
                domain=pyo.Reals,
                initialize=(annualization_factor * b_tec.para_unit_capex),
                mutable=True,
            )

        return b_tec

    def _define_capex_variables(self, b_tec, data: dict):
        """
        Defines variables related to storage technology capex for a flexible storage technology (with capex model 4).

        - var_capex_charging_cap: the capex of the charging capacity (MW) installed.
        - var_capex_discharging_cap: the capex of the discharging capacity (MW) installed.
        - var_capex_energy_cap: the capex of the energy storage capacity (MWh) installed.

        For a non-flexible storage technology (with fixed charging and discharging capacities), the capex
        variables are taken from the parent class (Technology).

        :param b_tec: pyomo block with technology model
        :param dict data: dict containing model information
        :return: pyomo block with technology model
        """

        super(Stor, self)._define_capex_variables(b_tec, data)

        config = data["config"]
        economics = self.economics
        capex_model = set_capex_model(config, economics)

        if capex_model == 4:
            coeff_ti = self.processed_coeff.time_independent

            # BOUNDS
            max_capex_charging_cap = (
                b_tec.para_unit_capex_charging_cap_annual
                * coeff_ti["charge_rate"]
                * b_tec.para_size_max
            )
            max_capex_discharging_cap = (
                b_tec.para_unit_capex_discharging_cap_annual
                * coeff_ti["discharge_rate"]
                * b_tec.para_size_max
            )
            max_capex_energy_cap = b_tec.para_unit_capex_annual * b_tec.para_size_max

            # VARIABLES
            b_tec.var_capex_charging_cap = pyo.Var(
                domain=pyo.NonNegativeReals, bounds=(0, max_capex_charging_cap)
            )
            b_tec.var_capex_discharging_cap = pyo.Var(
                domain=pyo.NonNegativeReals, bounds=(0, max_capex_discharging_cap)
            )
            b_tec.var_capex_energy_cap = pyo.Var(
                domain=pyo.NonNegativeReals, bounds=(0, max_capex_energy_cap)
            )

        return b_tec

    def _define_capex_constraints(self, b_tec, data: dict):
        """
        Defines constraints related to capex for a flexible storage technology of capex model 4, for which the capex
        comprises the capex for the charging capacity, the discharging capacity and the energy storage capacity.
        For a non-flexible storage technology, the capex constraints are taken from the parent class (Technology).

        :param b_tec: pyomo block with technology model
        :param dict data: dict containing model information
        :return: pyomo block with technology model
        """

        super(Stor, self)._define_capex_constraints(b_tec, data)

        config = data["config"]
        economics = self.economics
        capex_model = set_capex_model(config, economics)

        if capex_model == 4:

            # CAPEX constraints for flexible storage technologies
            # CAPEX_chargeCapacity = chargeCapacity * unitCost_chargeCapacity
            b_tec.const_capex_charging_cap = pyo.Constraint(
                expr=b_tec.var_capacity_charge
                * b_tec.para_unit_capex_charging_cap_annual
                == b_tec.var_capex_charging_cap
            )
            # CAPEX_dischargeCapacity = dischargeCapacity * unitCost_dischargeCapacity
            b_tec.const_capex_discharging_cap = pyo.Constraint(
                expr=b_tec.var_capacity_discharge
                * b_tec.para_unit_capex_discharging_cap_annual
                == b_tec.var_capex_discharging_cap
            )
            # CAPEX_storSize = storSize * unitCost_storSize
            b_tec.const_capex_energy_cap = pyo.Constraint(
                expr=b_tec.var_size * b_tec.para_unit_capex_annual
                == b_tec.var_capex_energy_cap
            )
            # CAPEX = CAPEX_chargeCapacity + CAPEX_dischargeCapacity + CAPEX_storSize
            b_tec.const_capex_aux = pyo.Constraint(
                expr=b_tec.var_capex_charging_cap
                + b_tec.var_capex_discharging_cap
                + b_tec.var_capex_energy_cap
                == b_tec.var_capex_aux
            )

        return b_tec

    def write_results_tec_design(self, h5_group, model_block):
        """
        Function to report technology design

        :param model_block: pyomo network block
        :param h5_group: h5 group to write to
        """
        super(Stor, self).write_results_tec_design(h5_group, model_block)

        if self.flexibility_data["power_energy_ratio"] == "flexratio":
            h5_group.create_dataset(
                "capacity_charge", data=[model_block.var_capacity_charge.value]
            )
            h5_group.create_dataset(
                "capacity_discharge", data=[model_block.var_capacity_discharge.value]
            )
            h5_group.create_dataset(
                "capex_charging_cap", data=[model_block.var_capex_charging_cap.value]
            )
            h5_group.create_dataset(
                "capex_discharging_cap",
                data=[model_block.var_capex_discharging_cap.value],
            )
            h5_group.create_dataset(
                "capex_energy_cap", data=[model_block.var_capex_energy_cap.value]
            )

    def write_results_tec_operation(self, h5_group, model_block):
        """
        Function to report technology operation

        :param model_block: pyomo network block
        :param h5_group: h5 group to write to
        """
        super(Stor, self).write_results_tec_operation(h5_group, model_block)

        h5_group.create_dataset(
            "storage_level",
            data=[model_block.var_storage_level[t].value for t in self.set_t_full],
        )

    def _define_ramping_rates(self, b_tec, data, sequence_storage):
        """
        Constraints the inputs for a ramping rate

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        dynamics = self.processed_coeff.dynamics

        ramping_time = dynamics["ramping_time"]

        # Calculate ramping rates
        if "ref_size" in dynamics and not dynamics["ref_size"] == -1:
            ramping_rate = dynamics["ref_size"] / ramping_time
        else:
            ramping_rate = b_tec.var_size / ramping_time

        # Constraints ramping rates
        if "ramping_const_int" in dynamics and dynamics["ramping_const_int"] == 1:

            s_indicators = range(0, 3)

            def init_ramping_operation_on(dis, t, ind):
                if t > 1:
                    if ind == 0:  # ramping constrained x[t] == x[t-1]
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == 0
                        )

                        def init_ramping_down_rate_operation_in(const):
                            # -rampingRate <= input[t] - input[t-1]
                            return (
                                -ramping_rate
                                <= self.input[
                                    t, self.component_options.main_input_carrier
                                ]
                                - self.input[
                                    t - 1, self.component_options.main_input_carrier
                                ]
                            )

                        dis.const_ramping_down_rate_in = pyo.Constraint(
                            rule=init_ramping_down_rate_operation_in
                        )

                        def init_ramping_up_rate_operation_in(const):
                            # input[t] - input[t-1] <= rampingRate
                            return (
                                self.input[t, self.component_options.main_input_carrier]
                                - self.input[
                                    t - 1, self.component_options.main_input_carrier
                                ]
                                <= ramping_rate
                            )

                        dis.const_ramping_up_rate_in = pyo.Constraint(
                            rule=init_ramping_up_rate_operation_in
                        )

                        def init_ramping_down_rate_operation_out(const):
                            # -rampingRate <= output[t] - output[t-1]

                            return -ramping_rate <= sum(
                                self.output[t, car_output]
                                - self.output[t - 1, car_output]
                                for car_output in b_tec.set_output_carriers
                            )

                        dis.const_ramping_down_rate_out = pyo.Constraint(
                            rule=init_ramping_down_rate_operation_out
                        )

                        def init_ramping_up_rate_operation_out(const):
                            # output[t] - output[t-1] <= rampingRate
                            return (
                                sum(
                                    self.output[t, car_output]
                                    - self.output[t - 1, car_output]
                                    for car_output in b_tec.set_output_carriers
                                )
                                <= ramping_rate
                            )

                        dis.const_ramping_up_rate_out = pyo.Constraint(
                            rule=init_ramping_up_rate_operation_out
                        )

                    elif ind == 1:  # startup, no ramping constraint x[t] - x[t-1] == 1
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == 1
                        )

                    else:  # shutdown, no ramping constraint x[t] - x[t-1] == -1
                        dis.const_ramping_on = pyo.Constraint(
                            expr=b_tec.var_x[t] - b_tec.var_x[t - 1] == -1
                        )

            b_tec.dis_ramping_operation_on = gdp.Disjunct(
                self.set_t_performance, s_indicators, rule=init_ramping_operation_on
            )

            # Bind disjuncts
            def bind_disjunctions(dis, t):
                return [b_tec.dis_ramping_operation_on[t, i] for i in s_indicators]

            b_tec.disjunction_ramping_operation_on = gdp.Disjunction(
                self.set_t_performance, rule=bind_disjunctions
            )

        else:
            if data["config"]["optimization"]["typicaldays"]["N"]["value"] == 0:
                input_aux_rr = self.input
                output_aux_rr = self.output
                set_t_rr = self.set_t_performance
            else:
                # init bounds at full res
                bounds_rr_full = {"input": {}, "output": {}}

                # Output Bounds
                for carr in self.component_options.output_carrier:
                    bounds_rr_full["output"][carr] = np.column_stack(
                        (
                            np.zeros(shape=(len(self.set_t_full))),
                            np.ones(shape=(len(self.set_t_full)))
                            * self.flexibility_data["discharge_rate"],
                        )
                    )

                # Input Bounds
                for carr in self.component_options.input_carrier:
                    if carr == self.component_options.main_input_carrier:
                        bounds_rr_full["input"][carr] = np.column_stack(
                            (
                                np.zeros(shape=(len(self.set_t_full))),
                                np.ones(shape=(len(self.set_t_full)))
                                * self.flexibility_data["charge_rate"],
                            )
                        )
                    else:
                        if (
                            "energy_consumption"
                            in self.input_parameters.performance_data["performance"]
                        ):
                            energy_consumption = self.input_parameters.performance_data[
                                "performance"
                            ]["energy_consumption"]
                            bounds_rr_full["input"][carr] = np.column_stack(
                                (
                                    np.zeros(shape=(len(self.set_t_full))),
                                    np.ones(shape=(len(self.set_t_full)))
                                    * self.flexibility_data["charge_rate"]
                                    * energy_consumption["in"][carr],
                                )
                            )

                # create input and output variable for full res
                def init_input_bounds(bounds, t, car):
                    return tuple(
                        bounds_rr_full["input"][car][t - 1, :]
                        * self.processed_coeff.time_independent["size_max"]
                        * self.processed_coeff.time_independent["rated_power"]
                    )

                def init_output_bounds(bounds, t, car):
                    return tuple(
                        bounds_rr_full["output"][car][t - 1, :]
                        * self.processed_coeff.time_independent["size_max"]
                        * self.processed_coeff.time_independent["rated_power"]
                    )

                b_tec.var_input_rr_full = pyo.Var(
                    self.set_t_full,
                    b_tec.set_input_carriers,
                    within=pyo.NonNegativeReals,
                    bounds=init_input_bounds,
                )
                b_tec.var_output_rr_full = pyo.Var(
                    self.set_t_full,
                    b_tec.set_output_carriers,
                    within=pyo.NonNegativeReals,
                    bounds=init_output_bounds,
                )

                b_tec.const_link_full_resolution_rr_input = (
                    link_full_resolution_to_clustered(
                        self.input,
                        b_tec.var_input_rr_full,
                        self.set_t_full,
                        self.sequence,
                        b_tec.set_input_carriers,
                    )
                )

                b_tec.const_link_full_resolution_rr_output = (
                    link_full_resolution_to_clustered(
                        self.output,
                        b_tec.var_output_rr_full,
                        self.set_t_full,
                        sequence_storage,
                        b_tec.set_output_carriers,
                    )
                )

                input_aux_rr = b_tec.var_input_rr_full
                output_aux_rr = b_tec.var_output_rr_full
                set_t_rr = self.set_t_full

            # Ramping constraint without integers
            def init_ramping_down_rate_input(const, t):
                # -rampingRate <= input[t] - input[t-1]
                if t > 1:
                    return (
                        -ramping_rate
                        <= input_aux_rr[t, self.component_options.main_input_carrier]
                        - input_aux_rr[t - 1, self.component_options.main_input_carrier]
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_down_rate = pyo.Constraint(
                set_t_rr, rule=init_ramping_down_rate_input
            )

            def init_ramping_up_rate_input(const, t):
                # input[t] - input[t-1] <= rampingRate
                if t > 1:
                    return (
                        input_aux_rr[t, self.component_options.main_input_carrier]
                        - input_aux_rr[t - 1, self.component_options.main_input_carrier]
                        <= ramping_rate
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_up_rate = pyo.Constraint(
                set_t_rr, rule=init_ramping_up_rate_input
            )

            def init_ramping_down_rate_output(const, t):
                # -rampingRate <= output[t] - output[t-1]
                if t > 1:
                    return -ramping_rate <= sum(
                        output_aux_rr[t, car_output] - output_aux_rr[t - 1, car_output]
                        for car_output in b_tec.set_output_carriers
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_down_rate_output = pyo.Constraint(
                self.set_t_performance, rule=init_ramping_down_rate_output
            )

            def init_ramping_down_rate_output(const, t):
                # output[t] - output[t-1] <= rampingRate
                if t > 1:
                    return (
                        sum(
                            output_aux_rr[t, car_output]
                            - output_aux_rr[t - 1, car_output]
                            for car_output in b_tec.set_output_carriers
                        )
                        <= ramping_rate
                    )
                else:
                    return pyo.Constraint.Skip

            b_tec.const_ramping_up_rate_output = pyo.Constraint(
                self.set_t_performance, rule=init_ramping_down_rate_output
            )

        return b_tec
